"""
Author(s):      Carson Schubert (carson.schubert14@gmail.com)  
Date Created:   03/13/2019

Utility module for managing Jigsaw created datasets.
"""

import json
import boto3
from botocore.exceptions import ClientError
from pathlib import Path
from ravenml.utils.local_cache import RMLCache
from ravenml.utils.config import get_config
from ravenml.utils.aws import list_top_level_bucket_prefixes, download_prefix
from ravenml.data.interfaces import Dataset

dataset_cache = RMLCache('datasets')
# name of dataset bucket field inside config dict
BUCKET_FIELD = 'dataset_bucket_name'

### PUBLIC METHODS ###
def get_dataset_names() -> list:
    """Retrieves the names of all available datasets in bucket pointed to by global config.

    Returns:
        list: dataset names
    """
    config = get_config()
    return list_top_level_bucket_prefixes(config[BUCKET_FIELD])

def get_dataset_metadata(name: str, no_check=False) -> dict:
    """Retrieves dataset metadata. Downloads from S3 if necessary.

    Args:
        name (str): string name of dataset
        no_check (bool, optional): whether to verify metadata existence

    Returns:
        dict: dataset metadata
        
    Raises:
        ValueError: if given dataset name is invalid (re raised)
    """
    if not no_check:
        try:
            _ensure_metadata(name)
        except ValueError:
            raise
    return json.load(open(dataset_cache.path / Path(name) / 'metadata.json'))

def get_dataset(name: str) -> Dataset:
    """Retrives a dataset. Downloads from S3 if necessary.

    Args:
        name (str): string name of dataset
    
    Returns:
        Dataset: dataset itself
        
    Raises:
        ValueError: if dataset name is invalid (re raised)
    """
    try:
        _ensure_dataset(name)
        return Dataset(name, get_dataset_metadata(name, no_check=True), dataset_cache.path / Path(name))
    except ValueError:
        raise
 

### PRIVATE HELPERS ###
def _ensure_metadata(name: str):
    """Ensure dataset metadata exists.

    Args:
        name (str): name of dataset
        
    Raises:
        ValueError: if dataset name is invalid and metadata cannot be downloaded.
    """
    S3 = boto3.resource('s3')
    config = get_config()
    DATASET_BUCKET = S3.Bucket(config['dataset_bucket_name'])
    metadata_path = Path(name) / 'metadata.json'
    if not dataset_cache.subpath_exists(metadata_path):
        dataset_cache.ensure_subpath_exists(name)
        metadata_key = f'{name}/metadata.json'
        metadata_absolute_path = dataset_cache.path / metadata_path
        try:
            DATASET_BUCKET.download_file(metadata_key, str(metadata_absolute_path))
        except ClientError as e:
            raise ValueError(name) from e

def _ensure_dataset(name: str):
    """Ensures dataset exists.

    Args:
        name (str): name of dataset
        
    Raises:
        ValueError: if dataset name is invalid (no matching objects in S3 bucket)
    """
    config = get_config()
    if not download_prefix(config[BUCKET_FIELD], name, dataset_cache):
        raise ValueError(name)



"""
Author(s):      Carson Schubert (carson.schubert14@gmail.com)  
Date Created:   02/23/2019

Main CLI entry point for ravenml.
"""

import click
from colorama import init, Fore
from ravenml.train.commands import train
from ravenml.data.commands import data
from ravenml.config.commands import config
from ravenml.utils.config import get_config, update_config
from ravenml.utils.local_cache import RMLCache

init()
cache = RMLCache()

### OPTIONS ###
clean_all_opt = click.option(
    '-a', '--all', is_flag=True,
    help='Clear all cache contents, including saved ravenML configuration.'
)


### COMMANDS ###
@click.group(help='Welcome to ravenML!')
def cli():
    """ Top level command group for ravenml.
    """
    pass
    
@cli.command(help='Cleans locally saved ravenML cache files.')
@clean_all_opt
def clean(all: bool):
    """ Cleans locally saved ravenml cache files.

    Args:
        all (bool): T/F whether to clean all files from cache, including
            configuration YAML, default false
    """
    if all:
        if not cache.clean():
            click.echo(Fore.RED + 'No cache to clean.')
    else:
        try:
            config = get_config()
            cache.clean()
            update_config(config)
        except ValueError:
            click.echo(Fore.RED + 'Bad configuration file. Deleting alongside cache.') 
            cache.clean()
        except FileNotFoundError:
            if not cache.clean():
                click.echo(Fore.RED + 'No cache to clean.')

cli.add_command(train)
cli.add_command(data)
cli.add_command(config)



import click

### HELPERS/CALLBACKS ###


### OPTIONS ###
"""Flag to determine if K-Fold Cross Validation will be used in training.
"""
dataset_name_opt = click.option(
    '--dataset-name', type=str,
    help='Name of dataset being created'
)

kfolds_opt = click.option(
    '--kfolds', '-k', type=int,
    help='Number of folds in dataset.'
)

upload_opt = click.option(
    '--upload', type=str,
    help='Enter the bucket you would like to upload too.'
)

delete_local_opt = click.option(
    '--delete-local', '-d', 'delete_local', is_flag=True,
    help='Delete local datasets.'
)

filter_opt = click.option(
    '-f', '--filter', is_flag=True, 
    help='Enable interactive image filtering option'
)

### PASS DECORATORS ###




"""
Author(s):      Carson Schubert (carson.schubert14@gmail.com)
Date Created:   04/05/2019

Utility module for managing ravenml's configuration.
"""

import yaml
from copy import deepcopy
from pathlib import Path
from ravenml.utils.local_cache import RMLCache

config_cache = RMLCache()

# required configuration fields
CONFIG_FIELDS = sorted(['image_bucket_name', 'dataset_bucket_name', 'model_bucket_name'])

def get_config() -> dict:
    """Retrieves the current configuration.
    
    Returns:
        dict: current configuration

    Raises:
        ValueError: If a required field is missing or an invalid field is found.
        FileNotFoundError: If a configuration file is not found.
    """
    config = {}
    if config_cache.subpath_exists('config.yml'):
        with open(config_cache.path / Path('config.yml'), 'r') as stream:
            try:
                config = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)
        required_fields = deepcopy(CONFIG_FIELDS)
        for key, value in config.items():
            if key in required_fields:    
                required_fields.remove(key)
            else:
                raise ValueError('Invalid field in configuration - ' + key)
        if len(required_fields) != 0:
            raise ValueError('Missing required configuration fields - ' + str(required_fields))
    else:
        raise FileNotFoundError('Configuration file does not exist.')
    return config

def update_config(config: dict):
    """Updates the configuration file.

    Args:
        config (dict): new configuration as a dict
    """
    # ensure our output location actually exists
    config_cache.ensure_exists()
    with open(config_cache.path / Path('config.yml'), 'w') as outfile:
        yaml.dump(config, outfile, default_flow_style=False)



"""
Author(s):      Carson Schubert (carson.schubert14@gmail.com)
Date Created:   03/19/2019

Classes necessary for interfacing with the data command group.
"""

import glob
from pathlib import Path

### CONSTANTS ###
# these should be used in all possible situations to protect us
# in case they change in the future
STANDARD_DIR = 'standard'
FOLD_DIR_PREFIX = 'fold_'
TEST_DIR = 'test'

# TODO add necessary functionality to this class as needed

class CreateInput(object):
    """Represents a dataset creation input. Contains all plugin-independent
    information necessary for training. Plugins can define their own behavior
    for getting additional information.

    """

class Dataset(object):
    """Represents a training dataset.

    Args:
        name (str): name of dataset 
        metadata (dict): metadata of dataset
        path (Path): filepath to dataset

    Attributes:
        name (str): name of the dataset 
        metadata (dict): metadata of dataset
        path (Path): filepath to dataset
    """
    def __init__(self, name: str, metadata: dict, path: Path):
        self.name = name
        self.metadata = metadata
        self.path = path
        
    def get_num_folds(self) -> int:
        """Gets the number of folds this dataset supports for 
        k-fold cross validation.

        Returns:
            int: number of folds
        """
        path = self.path / Path('dev')
        return len(glob.glob(str(path) + FOLD_DIR_PREFIX + '*'))
    



"""
Author(s):      Carson Schubert (carson.schubert14@gmail.com)  
Date Created:   4/11/2019

Standard options which should be used by ravenml plugins.

NOTE: there is currently nothing in this value except old functionality
that is no longer in use.
"""

import click

### HELPERS/CALLBACKS ###

# NOTE: the no_user callback and option are currently unused, however I am leaving it in the codebase
# so its functionality can be brought back if needed in the future. Its purpose is to
# ensure that ALL options to a command are set if the --no-user option is passed.
# def no_user_callback(ctx, param: click.core.Option, value: bool):
#     """Callback used by the no-user option. Evaluates all loaded parameters
#     and if any are still None, throws an error.

#     Args:
#         param (click.core.Option): option callback has been called by
#         value (bool): value of no-user option, default false
#     """
#     # set in context
#     if ctx.obj is None:
#         ctx.obj = {}
#     ctx.obj['NO_USER'] = value
#     # if we are in no_user mode, check all required arguments are there
#     if value:
#         for arg, value in ctx.params.items():
#             if value is None:
#                 # ctx.exit('You must supply the --%s argument when using --no-user!'%arg)
#                 raise click.exceptions.BadParameter('You must provide this argument when using --no-user!',
#                                                     ctx=ctx, param=arg, param_hint=str(arg))
# # NOTE: Any option intended to be used alongside the no_user_opt on a command
# # must have their is_eager flag set to TRUE for no_user_opt to work properly
# """Flag to determine if the command should run in user mode.
# """
# no_user_opt = click.option(
#     '--no-user', is_flag=True, callback=no_user_callback, expose_value=False,
#     help='Disable Inquirer prompts and use flags instead.'
# )

### OPTIONS ###

# NOTE: The pattern below is currently unused as we no longer make use of large
# numbers of options on a single command. I'm leaving it in the codebase in case its functionality
# is needed in the future. The option_decorator is a decorator that can be used on a click command
# using the same @option_decorator syntax as other decorators. It then applies all options
# defined in the "opts" list to that command in the order they are given in the list.
# This is just synactic sugar but makes the commands look much nicer.
# opts = [
#     verbose_opt,
#     comet_opt,
#     author_opt,
#     comments_opt,
#     model_name_opt,
#     optimizer_opt,
#     use_default_config_opt,
#     hyperparameters_opt
# ]

# ## Importable Option Decorator ##
# def option_decorator(func):
#     chain = opts[-1](func)
#     # must loop backwards so order of arguments in click
#     # command def is same as order in opts list
#     for i in range(len(opts) - 2, -1, -1):
#         chain = opts[i](chain)
#     return chain


### PASS DECORATORS ###
